/*
	PROforma, PROGS FOnt & Raster MAnager

	Font handling routines for PROforma.

	PFFontLoad
	PFFontUnLoad
	PFFontSelect
	PFFontScale
	PFFontNameGet
	PFFontWeightGet
	PFFontFamilyGet
	PFFontNoticeGet
	PFFontVersionGet
	PFFontCountAll
	PFFontCount
	PFFontNextName
	PFFontNext
	PFCacheUse
	PFFontCalculate

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 18 August 1993
	09.10.1995 modified naming scheme !!
	12.10.1995 changes to reflect better cache handling
			   changing the font nows remains the FontScale
	15.10.1995 support for resident fonts - there should always be at least one
	16.10.1995 removed some unnecessary tests for existence of currfont
			   PFFontSelect now selects builtin font on ERR_ITNF
	29.10.1995 PFFontNext & PFFontNextName can now return ERR_IPAR when end of
			   list reached, the counter now accept NULL as parameter
	09.11.1995 the error should better be ERR_EOF !
	28.12.1995 finally removed dependence on old _super and _user routines
	03.06.1996 width table now owned by PROforma, not loading job !!
	10.11.1996 currfontname now always set to allow restoring the font
			   PFFontSelect now a bit faster when selecting current font
	26.12.1996 PFFontCalculate added
	17.01.1997 - PFFontCount & PFFontNext now also cycle the resident fonts
			   - making font resident in PFFontCalculate now correct (usercount
				 was wrong)
*/

#include <io.h>
#include <str.h>
#include <mem.h>
#include <atomic.h>

#include "gstate.h"
#include "list.h"
#include "system.h"

/* some locally used/modified global variables */
font *fontlist=NULL;				/* list of loaded fonts */
font *ResFonts=NULL;				/* list of resident fonts */
extern fontline *fontmap;			/* list of available fonts */
extern char fontpath[100];			/* search path for font files */
extern Job PROforma;				/* PROforma job identificator */

/* some local utility routines */

/* routine to load a font */
static Error loadfont(const char *name, font **base)
{
	Error err;
	Channel chid;
	Size len;
	blockheader *bh;
	char *blockbase;
	font *f;
	font_width *widths;
	widthkey *currwidth;
	ref *currglyph;
	int i;

	/* open file */
	err=IOOpenPath(name,OPEN_READ,fontpath,&chid);
	if (err) return err;

	/* load image */
	err=IOLength(chid,&len);
	if (err)
	{
		IOClose(chid);
		return err;
	}
	err=MEMAllocateFor(len+offsetof(font,header),PROforma,base);
	if (err)
	{
		IOClose(chid);
		return err;
	}
	err=IOFileLoad(chid,len,(*base)->header);
	IOClose(chid);
	if (err)
	{
		MEMRelease(*base);
		*base=NULL;
		return err;
	}

	/* make sure the loaded file is indeed a PROforma font !! */
	f=*base;
	if(!MEMSame(26, f->header, "PROforma native font    01"))
	{
		MEMRelease(*base);
		*base=NULL;
		return ERR_ISYN;
	}

	/* fill in length and address of tables, rest is initialised to NULL */
	f->length=len;

	bh=&f->firstblock;
	blockbase=((char *)bh)+sizeof(blockheader);

	while (len>26)
	{
		switch(bh->type)
		{
		case font_information:
			f->info=(font_info *)blockbase;
			break;
		case font_subroutines:
			f->subrs=(font_routs *)blockbase;
			break;
		case font_glyphs:
			f->glyphs=(font_routs *)blockbase;
			break;
		case font_kerntable:
			f->kern=(font_kern *)blockbase;
			break;
		}
		/* now skip this block */
		len-=sizeof(blockheader)+bh->len;
		blockbase+=bh->len;
		bh=(blockheader *)blockbase;
		blockbase+=sizeof(blockheader);
	}

	/* a font needs at least its name and some outlines !! */
	if (f->info==NULL || f->glyphs==NULL)
	{
		MEMRelease(*base);
		*base=NULL;
		return ERR_ISYN;
	}

	/* I have to build the character width table for efficient PFWidth */
	err=PFAllocate(f->glyphs->count*sizeof(widthkey) +
				   offsetof(font_width,list), &widths);
	if (err)
	{
		MEMRelease(*base);
		*base=NULL;
		return ERR_IMEM;
	}
	f->widths=widths;
	i=widths->count=f->glyphs->count;
	for (currwidth=widths->list, currglyph=f->glyphs->list ;
		i ;
		i--, currwidth++, currglyph++)
	{
		int code;
		if (code=currglyph->code)
		{
			currwidth->unicode=code;
			currwidth->width=glyphwidth(f,currglyph->base+(int)f->glyphs);
			if (code>=PF_WIDTHARRAY_FIRST && code<=PF_WIDTHARRAY_LAST)
				widths->array[code-PF_WIDTHARRAY_FIRST]=currwidth->width;
		}
	}
	return err;
}

static Error linkres(font *resident)
{
	font *cf;

	resident->usercount=-1; 		/* indicate that this font is resident */
	if (ResFonts==NULL)
		ResFonts=resident;
	else
	{
		for (cf=ResFonts ; cf->next ; cf=cf->next);
		cf->next=resident;
	}
	return ERR_OK;
}

/* add a resident font in PROforma */
Error FontResidentLoad(char *filename)
{
	Error err;
	font *resident;

	err=loadfont(filename, &resident);
	if (err) return err;
	ATOMICCall(linkres,1,resident);
	return err;
}


/* scale the current font to the given size */
Error PFFontScale(Gstate gstate, pt scale)
{
	if (gstate->fontsize!=scale)
	{
		CacheFontRelease(&gstate->font);
		gstate->fontsize=scale;
	}
	return ERR_OK;
}

/* aid routine to select a font, fontsize is not affected by this */
static Error selectfont(Gstate gstate, font *f)
{
	if (gstate->currfont!=f)
	{
		gstate->currfont=f;
		STRCopy(gstate->currfont->info->FullName,gstate->currfontname);
		CacheFontRelease(&gstate->font);
	}
	return ERR_OK;
}

Error PFFontSelect(Gstate gstate, const char *fontname)
{
	fontrec *fr=gstate->fontlist;

	if (STRSameCD(gstate->currfont->info->FullName,fontname))
		return ERR_OK;

	while (fr && !STRSameCD(fr->FullName,fontname)) fr=fr->next;
	if (fr)
	{
		gstate->currfont=fr->base;
		STRCopy(gstate->currfont->info->FullName,gstate->currfontname);
		CacheFontRelease(&gstate->font);
		return ERR_OK;
	}
	else
	{
		/* see if there is a resident font which we can use */
		font *f;
		for (f=ResFonts ; f ; f=f->next)
		{
			if (STRSameCD(f->info->FullName,fontname))
			{
				gstate->currfont=f;
				STRCopy(gstate->currfont->info->FullName,gstate->currfontname);
				CacheFontRelease(&gstate->font);
				return ERR_OK;
			}
		}
	}
	/* not found -- select "builtin" font */
	gstate->currfont=ResFonts;
	STRCopy(gstate->currfont->info->FullName,gstate->currfontname);
	CacheFontRelease(&gstate->font);
	if (STRSameCD(gstate->currfont->info->FullName,fontname)) return ERR_OK;
	return ERR_ITNF;
}

static void unloadfont(void *base)
{
	fontrec *fr=(fontrec *)base;

	/* release this fontrec and possibly the font */
	fr->base->usercount--;
	if (fr->base->usercount==0)
	{
		font *f=fontlist, *f2;
		if (fontlist==fr->base)
			fontlist=fr->base->next;
		else
		{
			while (f->next!=fr->base) f=f->next;
			f2=f;
			f=f->next;
			f2->next=f->next;
		}
		FontReleaseCaches(f);
		PFRelease(f->widths);
		MEMRelease(f);
	}
}

/* actually search for a font in the fontlist - has to be called atomically */
int searchfont(char *fontname, font **which)
{
	int found=FALSE;
	font *f=fontlist;

	while (f && !found)
	{
		if (STRSameCD(f->info->FullName,fontname))
		{
			f->usercount++;
			found=TRUE;
		}
		else
			f=f->next;
	}
	*which=f;
	return found;
}

Error fontlink(font *f)
{
	f->next=fontlist;
	fontlist=f;
	return ERR_OK;
}

/*	PFFontLoad
	loads a font file into memory, needs a fontname
	if the fontfile can't be opened, then the configured path(s) are searched
	an error is returned if the font type is invalid, or too long for memory
						 of not found
	the font which is loaded is automatically selected */
Error PFFontLoad(Gstate gstate, const char *fontname)
{
	font *f;
	fontrec *fr;
	char filename[IO_MAXFULLNAME];
	Size len; char *base; blockheader *bh; char found=FALSE;
	fontline *fl=fontmap;
	Error err;

	/* first see if this font is already in use by this Gstate */
	err=PFFontSelect(gstate, fontname);
	if (!err) return err;

	/* create a fontrec for this new font */
	err=MEMAllocate(sizeof(fontrec),&fr);
	if (err) return err;

	/* search font : if already in memory -> just one more user */
	if (!ATOMICCall(searchfont,2,fontname,&f))
	{
		/* ok, load the font */

		/* we have to find the filename first */
		while (fl && !STRSameCD(fl->FullName, fontname)) fl=fl->next;
		if (fl)
			STRCopy(fl->filename, filename);
		else
		{
			MEMRelease(fr);
			return ERR_IEXP;
		}

		err=loadfont(filename,&f);					/* load fontfile */
		if (err) { MEMRelease(fr); return err; }    /* oops, ERROR */

		f->usercount=1;
		ATOMICCall(fontlink,1,f);
	}
	CacheFontRelease(&gstate->font);
	gstate->currfont=f;
	STRCopy(gstate->currfont->info->FullName,gstate->currfontname);
	fr->base=f;
	MEMWhenZapped(fr,unloadfont,NULL);
	MEMCopy(strlen(gstate->currfont->info->FullName)+1,
				   gstate->currfont->info->FullName, fr->FullName);
	linkin(fr,gstate->fontlist);
	return ERR_OK;
}

/* PFFontUnLoad */
/* make sure this Gstate can no longer use the font (current font=builtin) */
Error PFFontUnLoad(Gstate gstate, const char *fontname)
{
	fontrec *fr=gstate->fontlist, *fr2;

	/* see whether this font exists */
	if (PFFontSelect(gstate, fontname)) return ERR_OK;
	/* make sure it is not a resident font */
	if (gstate->currfont->usercount==-1) return ERR_OK;

	/* we have to release the fontrec */
	if (fr->base==gstate->currfont)
		gstate->fontlist=fr->next;
	else
	{
		while (fr->next->base!=gstate->currfont) fr=fr->next;
		fr2=fr;
		fr=fr->next;
		fr2->next=fr->next;
	}

	MEMRelease(fr);

	/* make current font undefined (meaning builtin) */
	gstate->currfont=ResFonts;
	STRCopy(gstate->currfont->info->FullName,gstate->currfontname);
	return ERR_OK;
}

/* PFFontNameGet, get name of currently selected font */
Error PFFontNameGet(Gstate gstate, char *name)
{
	MEMCopy(PF_MAXFONTNAME,gstate->currfont->info->FullName,name);
	return ERR_OK;
}

/* PFFontWeightGet, get weight of currently selected font */
Error PFFontWeightGet(Gstate gstate, char *name)
{
	MEMCopy(PF_MAXFONTWEIGHT,gstate->currfont->info->Weight,name);
	return ERR_OK;
}

/* PFFontFamilyGet, get FamilyName of currently selected font */
Error PFFontFamilyGet(Gstate gstate, char *name)
{
	MEMCopy(PF_MAXFONTFAMILY,gstate->currfont->info->FamilyName,name);
	return ERR_OK;
}

/* PFFontNoticeGet, get Notice for currently selected font */
Error PFFontNoticeGet(Gstate gstate, char *name)
{
	MEMCopy(PF_MAXFONTNOTICE,gstate->currfont->info->Notice,name);
	return ERR_OK;
}

/* PFFontVersionGet, get version of currently selected font */
Error PFFontVersionGet(Gstate gstate, char *name)
{
	MEMCopy(PF_MAXFONTVERSION,gstate->currfont->info->Version,name);
	return ERR_OK;
}

/* get number of fonts in fontmap */
Error PFFontCountAll(Gstate gstate, int *count)
{
	fontline *fl=fontmap;
	int _count;

	if (count==NULL) count=&_count;

	*count=0;
	while (fl)
	{
		(*count)++;
		fl=fl->next;
	}
	gstate->nextfont=fontmap;
	return ERR_OK;
}

/* get number of loaded fonts, resets current font to first font */
Error PFFontCount(Gstate gstate, int *count)
{
	fontrec *fr;
	font *f;
	int _count;

	if (count==NULL) count=&_count;

	*count=0;
	for (fr=gstate->fontlist ; fr ; fr=fr->next) (*count)++;
	for (f=ResFonts ; f ; f=f->next) (*count)++;
	return selectfont(gstate,ResFonts);
}

/* get name of next font in fontmap */
Error PFFontNextName(Gstate gstate, char *name)
{
	name[0]='\0';
	if (gstate->nextfont)
	{
		STRCopy(gstate->nextfont->FullName, name);
		gstate->nextfont=gstate->nextfont->next;
		return ERR_OK;
	}
	else
		return ERR_EOF;
}

/* select next font in list of loaded fonts */
Error PFFontNext(Gstate gstate)
{
	fontrec *fr=gstate->fontlist;
	font *f=ResFonts;

	for (f=ResFonts ; f && f!=gstate->currfont ; f=f->next);
	if (f==gstate->currfont)
	{
		if (!f->next && !fr) return ERR_EOF;
		selectfont(gstate,f->next?f->next:fr->base);
		return ERR_OK;
	}
	while (fr && fr->base!=gstate->currfont) fr=fr->next;
	if (fr && fr->next)
	{
		selectfont(gstate, fr->next->base);
		return ERR_OK;
	}
	return ERR_EOF;
}

/* Switch font cache usage on and/or off */
Error PFCacheUse(Gstate gstate, int status)
{
	gstate->usecache=status;
	return ERR_OK;
}

static Error makeresident(font *f)
{
	font *fl=fontlist, *trail=NULL;

	f->usercount=-1;
	for ( ; fl && fl!=f ; trail=fl, fl=fl->next);
	if (fl)
	{
		if (trail)
			trail->next=f->next;
		else
			fontlist=f->next;
		for (fl=ResFonts ; fl->next ; fl=fl->next);
		fl->next=f;
	}
	return ERR_OK;
}

Error PFFontCalculate(Gstate gstate)
{
	short i;

	if (!gstate->font) CacheFontFind(gstate);
	CacheFontCalculate(gstate->font);

	/* font should now be treated as a resident font */
	ATOMICCall(makeresident,1,gstate->currfont);

	gstate->noDisplayChars=TRUE;
	PFShowChar(gstate,'m'); 		/* for fixed width fonts */
	for (i=0x20 ; i<0x7f ; i++)
		PFShowChar(gstate,i);
	gstate->noDisplayChars=FALSE;
	return ERR_OK;
}



	
