/*
	PROforma, PROgs FOnt & Raster MAnager

	PFWindowSubSet
	PFWindowSubRestore
		Routines to set and restore subwindows. This concept can be used
		recursive. A SubWindow is somewhat similar to the PageBbox, however,
		the pagebbox is used to redraw part of something, while the SubWindow
		is used to draw the contents of a part of the gstate, as a subwindow
		on the screen. This is intended for screen gstates only. Any
		PageBbox has to fall inside the current SubWindow. Any successive Sub-
		Windows also has to fall inside the current SubWindow.
		Any call of PFPageScale is local inside the current and subsequent
		SubWindows.
		The new SubWindow is always relative with the currect PageBbox, and
		the coordinates are in user space (which should not be rotated).
		When the SubWindow is set, the PageOrigin is (0,0)
	PFPWindowMove
		To allow the window to be moved for a screen Gstate...
	PFBufferUpdate
		Update the internal buffer which is used by PROforma to display what
		is visible on screen.
	PFBufferSave, PFBufferRestore
		Save and restore the buffer to an internal copy. This gives extra
		support for interactive applications.
	PFPWindowLink
		Routine to link a screen save area to a window. To make it possible to
		assign a screen gstate to a part of a window or to a window which is
		too small. The window is controlled by the client and not the screen
		driver !

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 16 May 1995
	09.10.1995 naming scheme changed
	12.10.1995 some smaller changes for different cache
	14.10.1995 tried not to release gstate->font on PFWindowSubRestore
			   PFWindowSubRestore no longer releases the memory but puts it in
			   a list of empty boxes for later reuse - should be faster
	10.11.1995 PFWindowSubSet now also allowed for multi pass gstates.
	14.12.1995 PFBufferUpdate added
	23.12.1995 PFBufferSave and PFBufferRestore added
	20.03.1996 PFPWindowSubSet added
	29.03.1996 PFPWindowSubSet now takes PageOrigin into account
	10.11.1996 PFWindowSubRestore did not restore CTM properly
	08.10.1997 PFWindowSubSet switched some parameters
	16.11.1997 PFWindowSubRestore could keep old CTM when recovering from a
			   different PFPageScale. This is now fixed.
	23.12.1997 PFPWindowLink added
*/

#include "gstate.h"
#include "system.h"
#include "list.h"

Error PFWindowSubRestore(Gstate gstate)
{
	boxlist *b;
	int nb=FALSE;

	if (gstate->subwindows==NULL) return ERR_ITNF;
	linkout(b,gstate->subwindows);

	gstate->pagebbox	=b->pagebbox;
	gstate->raster->clip=b->clip;
	gstate->Origin		=b->Origin;
	if (gstate->todev.x!=b->todev.x)
	{
		nb=TRUE;
		gstate->ppp 		=b->ppp;
		gstate->todev		=b->todev;
	}
	gstate->Scale		=b->Scale;

	/* currCTM should also convert to device coordinates !! */
	if (nb || !MEMSame(sizeof(CTM),&b->todefaultCTM,&gstate->todefaultCTM))
	{
		MEMCopy(sizeof(CTM),&b->todefaultCTM,&gstate->todefaultCTM);
		currCTMset(gstate);
	}

	linkin(b,gstate->freesubwins);

	return ERR_OK;
}

static Error WindowSubSet(Gstate gstate, short xsiz, short ysiz,
										 short xorg, short yorg)
{
	Error err;
	boxlist *bl;
	PPageBbox *b=&gstate->pagebbox;
	PBbox *clip=&gstate->raster->clip;
	coor c;

	if (gstate->currCTM.xr || gstate->currCTM.yr) return ERR_IPAR;

	if (gstate->freesubwins)
	{
		linkout(bl,gstate->freesubwins);
	}
	else
	{
		err=MEMAllocate(sizeof(boxlist),&bl);
		if (err) return err;
	}

	bl->next		=NULL;
	bl->todefaultCTM=gstate->todefaultCTM;
	bl->pagebbox	=gstate->pagebbox;
	bl->clip		=gstate->raster->clip;
	bl->Origin		=gstate->Origin;
	bl->ppp 		=gstate->ppp;
	bl->todev		=gstate->todev;
	bl->Scale		=gstate->Scale;
	linkin(bl,gstate->subwindows);

	/* The PageBbox is adjusted, relative to the previous values */
	/* The given coordinates are in pixles */

	/* adjust PageOrigin */
	b->xporg=b->yporg=0;

	/* make coordinates absolute */
	xorg+=clip->xorg;
	yorg+=clip->yorg;

	/* now test whether the given values are in range */
	/* and clip them if they are not */
	if (xorg < clip->xorg)
	{
		short diff=clip->xorg-xorg;
		xorg+=diff;
		xsiz-=diff;
		b->xporg+=short2pt(diff);
	}
	if (yorg < clip->yorg)
	{
		short diff=clip->yorg-yorg;
		yorg+=diff;
		ysiz-=diff;
		b->yporg+=short2pt(diff);
	}
	if (xsiz>clip->xsiz+clip->xorg-xorg)
		xsiz=clip->xsiz+clip->xorg-xorg;
	if (ysiz>clip->ysiz+clip->yorg-yorg)
		ysiz=clip->ysiz+clip->yorg-yorg;

	/* now set the new values */
	b->xsiz=xsiz;
	b->ysiz=ysiz;
	b->xorg=xorg;
	b->yorg=yorg;

	PFPageBboxReset(gstate);

	return ERR_OK;
}

Error PFWindowSubSet(Gstate gstate, pt xsiz, pt ysiz, pt xorg, pt yorg)
{
	Error err;
	coor c;

	if (gstate->currCTM.xr || gstate->currCTM.yr) return ERR_IPAR;

	/* The PageBbox is adjusted, relative to the previous values */
	/* The given coordinates are in user space, BUT the CTM should only */
	/* consist of moves and scales - or error */

	/* calculate the new values */
	c.x=xorg; c.y=yorg;
	user2dev(gstate,&c);
	xorg=c.x; yorg=c.y;
	c.x=xsiz; c.y=ysiz;
	reluser2dev(gstate,&c);
	xsiz=c.x; ysiz=c.y;

	/* we always need some kind of window !! */
	if (xsiz<=pt_half || ysiz<=pt_half) { xsiz=0; ysiz=0; }

	/* make coordinates absolute and integer */
	return WindowSubSet(gstate, pt2rshort(xsiz), pt2rshort(ysiz),
								pt2rshort(xorg), pt2rshort(yorg));
}

Error PFPWindowSubSet(Gstate gstate, short xsiz, short ysiz,
									 short xorg, short yorg)
{
	register coor *porg=&gstate->Origin;
	return WindowSubSet(gstate, xsiz,ysiz, xorg-pt2rshort(porg->x),
										   yorg-pt2rshort(porg->y));
}

/* this routine should be called for screen Gstates of which the window
   may have been moved, distance in pixels */
Error PFPWindowMove(Gstate gstate, short dx, short dy)
{
	return gstate->printer->WindowMove(gstate, dx,dy);
}

Error PFBufferUpdate(Gstate gstate)
{
	return gstate->printer->BufferUpdate(gstate);
}

Error PFBufferSave(Gstate gstate)
{
	return gstate->printer->BufferSave(gstate);
}

Error PFBufferRestore(Gstate gstate)
{
	return gstate->printer->BufferRestore(gstate);
}

Error PFPWindowLink(Gstate gstate, Channel chid, short wxsiz, short wysiz,
					short wxorg, short wyorg, short bxorg, short byorg)
{
	return gstate->printer->Handle(2,gstate,chid,wxsiz,wysiz,
												 wxorg,wyorg,bxorg,byorg);
}
